package com.bayread.sensors.batch;

import com.bayread.sensors.batch.impl.SingleWorkerEvent;
import com.bayread.sensors.batch.impl.WorkerEvent;
import com.bayread.sensors.batch.impl.MultipleWorkerEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.Closeable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * 批量异步处理队列
 * 
 */
public class BatchWorkerCollector<T> implements WorkerCollector<T>{
	/**
	 * 服务名称
	 */
	private String name;
	
	/**
	 * 数据队列
	 */
	private BlockingQueue<T> bq;
	
	/**
	 * 批量数据队列
	 */
	private BlockingQueue<Runnable> batchQueue;
	
	/**
	 * 处理逻辑实现
	 */
	private WorkerEvent event;
	
	/**
	 * 消费等待时间(毫秒)
	 */
	private int pollWaitTime;
	
	/**
	 * 处理一批队列数据的大小
	 */
	private int batchSize;
	
	/**
	 * 处理线程数量
	 */
	private int threadSize;
	
	/**
	 * 指任务队列大小
	 */
	private int workQueueSize;

	/**
	 * 是否开启消费服务
	 */
	private volatile boolean isStart = false;

	private boolean isBatch = true;

	private Logger logger = LoggerFactory.getLogger(this.getClass());


	public BatchWorkerCollector(String name, int queueSize, int workQueueSize, int threadSize, int batchSize, int pollWaitTime, WorkerEvent event){
		this.name = name;
		this.bq = new LinkedBlockingQueue<T>(queueSize);
		this.batchQueue =  new LinkedBlockingQueue<Runnable>(workQueueSize<=0?Integer.MAX_VALUE:workQueueSize);
		this.event = event;
		this.workQueueSize = workQueueSize;
		this.batchSize = batchSize;
		this.pollWaitTime = pollWaitTime;
		this.threadSize = threadSize;
		this.isBatch = !(event instanceof SingleWorkerEvent);

		logger.info("BATCH QUEUE CONFIG[name:{} queuesize:{} threadPoolQueueSize:{} threadSize:{} batchSize:{}]",name,queueSize,workQueueSize,threadSize,pollWaitTime);
	}
	
	/**
	 * 开启消费服务
	 */
	public void start(){
		isStart = true;
		new Thread(new HandleThread(new InnerThreadPool(name, this.threadSize, workQueueSize))).start();
		logger.info("[{}]服务已开启.",name);
	}
	
	/**
	 * 关闭消费服务
	 */
	public void stop(){
		isStart = false;
		if ( this.event instanceof Closeable) {
			try {
				((Closeable)this.event).close();
			} catch (IOException e) {}
		}
	}

	/**
	 * 如果超过设置的队列阀值,将会阻塞等待
	 */
	public void put(T obj) {
		try {
			bq.put(obj);
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * 如果超过设置的队列阀值,将会直接丢弃并反回FALSE
	 */
	public boolean offer(T obj) {
		if(null == obj)return false;
		return bq.offer(obj);
	}
	
	/**
	 * 队列是否为空
	 */
	public boolean isEmpty(){
		return bq.isEmpty();
	}
	
	/**
	 * 队列大小
	 */
	public int size(){
		return bq.size();
	}
	
	/**
	 * 设置消费等待时间(毫秒)
	 */
	public void setPollWaitTime(int pollWaitTime) {
		this.pollWaitTime = pollWaitTime;
	}

	/**
	 * 设置处理一批队列数据的大小
	 */
	public void setBatchSize(int batchSize) {
		this.batchSize = batchSize;
	}
	
	/**
	 * 设置处理线程的数量
	 */
	public void setThreadSize(int threadSize) {
		this.threadSize = threadSize;
	}

	/**
	 * 设置线程池任务队列的大小
	 */
	public void setWorkQueueSize(int workQueueSize) {
		this.workQueueSize = workQueueSize;
	}
	
	/**
	 * 批量队列数量
	 */
	@Override
	public int batchQueueSize() {
		return batchQueue.size();
	}
	
	/**
	 * 队列消费线程
	 */
	final private class HandleThread implements Runnable{
		private ThreadPoolExecutor executor;
		
		private HandleThread(ThreadPoolExecutor executor){
			this.executor = executor;
		}
		
		@Override
		public void run() {
			while(isStart){
				try {
					if(isBatch){
						List<T> li = new ArrayList<T>(batchSize);

						for(int i=0;i<batchSize;i++){
							T t = bq.poll(pollWaitTime, TimeUnit.MILLISECONDS);
							if(null == t){break;}
							li.add(t);
						}

						if(li.isEmpty()){
							continue;
						}

						executor.execute(new Thread(){
							public void run(){
								try {
									((MultipleWorkerEvent)event).handle(li);
								} catch (Exception e) {
									e.printStackTrace();
									event.handleException(li,e);
									logger.error("[{}]批量队列任务执行异常.",name,e);
								}
							}
						});
					}else{
						T l = bq.poll(pollWaitTime, TimeUnit.MILLISECONDS);
						try {
							if(l != null) {
								((SingleWorkerEvent) event).handle(l);
							}
						} catch (Exception e) {
							event.handleException(l,e);
							logger.error("[{}]批量队列任务执行异常.",name,e);
						}
					}


				} catch (Exception e) {
					logger.error("[{}]线程任务分配异常.",name,e);
				}
			}
			
			logger.info("[{}]服务已关闭.",name);
			
			executor.shutdown();	
		}
	}
	
	/**
	 * 线程池
	 * 
	 */
	final private class InnerThreadPool extends ThreadPoolExecutor{
		public InnerThreadPool(String name,int threadCount,int taskQueueSize) {
			super(threadCount, (int) (threadCount * 1.618), 60, TimeUnit.MILLISECONDS,batchQueue,new InnerThreadFactory(name),new CallerRunsPolicy());
		}
	}
	
	/**
	 * 提供线程的工厂
	 * 
	 */
	final private class InnerThreadFactory implements ThreadFactory {
		private final ThreadGroup group;
		private final AtomicInteger threadNumber = new AtomicInteger(1);
		private final String namePrefix;

		InnerThreadFactory(String name) {
			SecurityManager s = System.getSecurityManager();
			group = (s != null) ? s.getThreadGroup() : Thread.currentThread().getThreadGroup();
			namePrefix = name + "-";
		}

		public Thread newThread(Runnable r) {
			Thread t = new Thread(group, r, namePrefix + threadNumber.getAndIncrement(), 0);
			if (t.isDaemon())
				t.setDaemon(false);
			if (t.getPriority() != Thread.NORM_PRIORITY)
				t.setPriority(Thread.NORM_PRIORITY);
			return t;
		}
	}
	
	public int getBatchSize() {
		return batchSize;
	}

	public int getPollWaitTime() {
		return pollWaitTime;
	}
	
	public int getThreadSize() {
		return threadSize;
	}

	public int getWorkQueueSize() {
		return workQueueSize;
	}
}